/*

MP3 bitrate grapher

produces bitrate histogram of MPEG Audio files for gnuplot
Copyright 2007 tmkk
Inspired by vinfo by Aoyumi

*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>


#ifdef _BIG_ENDIAN
#define SWAP32(n) (n)
#define SWAP16(n) (n)
#else
#define SWAP32(n) (((n>>24)&0xff) | ((n>>8)&0xff00) | ((n<<8)&0xff0000) | ((n<<24)&0xff000000))
#define SWAP16(n) (((n>>8)&0xff) | ((n<<8)&0xff00))
#endif

typedef struct {
    unsigned char id[3];
    unsigned short version;
    unsigned char flags;
    unsigned char size[4];
} __attribute__ ((packed)) id3v2_t;

int calc_bitrate(int header, int freq)
{
	if((header >> 21) & 0x7ff != 0x7ff) return 0;
	int rate = (header >> 12) & 0xf;
	int padding = (header >> 9) & 0x1;
	
	if(freq == 32000) {
		switch(rate) {
		  case 0x1: return 144+padding;
		  case 0x2: return 180+padding;
		  case 0x3: return 216+padding;
		  case 0x4: return 252+padding;
		  case 0x5: return 288+padding;
		  case 0x6: return 360+padding;
		  case 0x7: return 432+padding;
		  case 0x8: return 504+padding;
		  case 0x9: return 576+padding;
		  case 0xa: return 720+padding;
		  case 0xb: return 864+padding;
		  case 0xc: return 1008+padding;
		  case 0xd: return 1152+padding;
		  case 0xe: return 1440+padding;
		  default: return 0;
		}
	}
	else if(freq == 44100) {
		switch(rate) {
		  case 0x1: return 104+padding;
		  case 0x2: return 130+padding;
		  case 0x3: return 156+padding;
		  case 0x4: return 182+padding;
		  case 0x5: return 208+padding;
		  case 0x6: return 261+padding;
		  case 0x7: return 313+padding;
		  case 0x8: return 365+padding;
		  case 0x9: return 417+padding;
		  case 0xa: return 522+padding;
		  case 0xb: return 626+padding;
		  case 0xc: return 731+padding;
		  case 0xd: return 835+padding;
		  case 0xe: return 1044+padding;
		  default: return 0;
		}
	}
	else if(freq == 48000) {
		switch(rate) {
		  case 0x1: return 96+padding;
		  case 0x2: return 120+padding;
		  case 0x3: return 144+padding;
		  case 0x4: return 168+padding;
		  case 0x5: return 192+padding;
		  case 0x6: return 240+padding;
		  case 0x7: return 288+padding;
		  case 0x8: return 336+padding;
		  case 0x9: return 384+padding;
		  case 0xa: return 480+padding;
		  case 0xb: return 576+padding;
		  case 0xc: return 672+padding;
		  case 0xd: return 768+padding;
		  case 0xe: return 860+padding;
		  default: return 0;
		}
	}
	return 0;
}

int calc_freq(int header)
{
	if((header >> 21) & 0x7ff != 0x7ff) return 0;
	int freq = (header >> 10) & 0x3;
	int version = (header >> 19) & 0x3;
	
	if(version == 3) {
		switch(freq) {
		  case 0: return 44100;
		  case 1: return 48000;
		  case 2: return 32000;
		  default: return 0;
		}
	}
	else if(version == 2) {
		switch(freq) {
		  case 0: return 22050;
		  case 1: return 44000;
		  case 2: return 16000;
		  default: return 0;
		}
	}
	else if(version == 0) {
		switch(freq) {
		  case 0: return 11025;
		  case 1: return 12000;
		  case 2: return 8000;
		  default: return 0;
		}
	}
	return 0;
}

int calc_spf(int header)
{
	int layer = (header >> 17) & 0x3;
	int version = (header >> 19) & 0x3;
	if(layer == 2 || (version == 3 && layer == 1)) { //layer 2, MPEG1 layer 3
		return 1152;
	}
	else if(layer == 3) { //layer 1
		return 384;
	}
	else if(layer == 1 && (version == 2 || version == 0)) { //MPEG2/2.5 layer 3
		return 576;
	}
	
	return 0;
}

int main(int argc, char *argv[])
{
	FILE *fp,*fpw1,*fpw2;
	int i,tmp,time,frameNum,sample,size=0,bitrate,resolution,samplePerFrame;
	double frameSize;
	char atom[4],filename[1024];
	char *ptr;
	id3v2_t id3v2;
	
	if(argc<2 || argc > 3) {
		fprintf(stderr,"usage: %s mp3file.mp3 [resolution]\n",argv[0]);
		return 0;
	}
	
	for(ptr=argv[1]+strlen(argv[1]);ptr!=argv[1];ptr--) {
		if(*ptr == '/') {
			ptr++;
			break;
		}
	}
	
	fp = fopen(argv[1], "rb");
	sprintf(filename,"%s_bitrate.plt",ptr);
	fpw1 = fopen(filename,"wb");
	sprintf(filename,"%s_bitrate.txt",ptr);
	fpw2 = fopen(filename,"wb");
	if(!fpw1 || !fpw2) {
		fprintf(stderr,"can't write bitrate file\n");
		goto end;
	}
	
	getcwd(filename,1024);
	sprintf(filename,"%s/%s_bitrate.txt",filename,ptr);
	
	if (fread(&id3v2, sizeof(id3v2), 1, fp) == 0) goto end;
	
	if (!memcmp(id3v2.id, "ID3", 3)) {
		long len;

		if (id3v2.size[0] & 0x80) goto end;

		len = (id3v2.size[0] & 0x7f);
		len = (len << 7) | (id3v2.size[1] & 0x7f);
		len = (len << 7) | (id3v2.size[2] & 0x7f);
		len = (len << 7) | (id3v2.size[3] & 0x7f);
		len += 10;
		if (id3v2.flags & (1 << 4)) len += 10;

		if(fseek(fp, len, SEEK_SET) != 0) goto end;
	} else if(fseek(fp, 0, SEEK_SET) != 0) goto end;
	
	if(fread(&tmp,4,1,fp) < 1) goto end;
	int frequency = calc_freq(SWAP32(tmp));
	if(!frequency) {
		fprintf(stderr,"error: seems to be not an MP3 file...\n");
		goto end;
	}
	
	samplePerFrame = calc_spf(SWAP32(tmp));
	if(!samplePerFrame) {
		fprintf(stderr,"error: seems to be not an MP3 file...\n");
		goto end;
	}
	
	//printf("%d,%d\n",frequency,samplePerFrame);
	
	if(argc == 3 && atoi(argv[2]) >= samplePerFrame) resolution = atoi(argv[2]);
	else resolution = frequency;
	
	if(fseek(fp, -4, SEEK_CUR) != 0) goto end;
	
	if(resolution == frequency) {
		for(i=0,sample=0,frameSize=0,time=1;;i++) {
			if(fread(&tmp,4,1,fp) < 1) goto last;
			tmp = calc_bitrate(SWAP32(tmp),frequency);
			if(!tmp) goto last;
			size += tmp;
			if(sample + samplePerFrame >= frequency) {
				frameSize += (double)(frequency - sample)/samplePerFrame * tmp;
				fprintf(fpw2,"%02d:%02d\t%f\n",time/60, time-60*(time/60), frameSize *8);
				sample = samplePerFrame - (frequency - sample);
				frameSize = (double)sample/samplePerFrame * tmp;
				time++;
				if(fseek(fp,tmp-4,SEEK_CUR) != 0) goto last;
				continue;
			}
			frameSize += tmp;
			sample += samplePerFrame;
			if(fseek(fp,tmp-4,SEEK_CUR) != 0) goto last;
		}
	}
	else {
		for(i=0,sample=0,frameSize=0,time=1;;i++) {
			if(fread(&tmp,4,1,fp) < 1) goto last;
			tmp = calc_bitrate(SWAP32(tmp),frequency);
			if(!tmp) goto last;
			size += tmp;
			if(sample + samplePerFrame >= resolution) {
				frameSize += (double)(resolution - sample)/samplePerFrame * tmp;
				fprintf(fpw2,"%f\t%f\n",resolution*time/(double)frequency, frameSize * ((double)frequency/resolution) * 8);
				sample = samplePerFrame - (resolution - sample);
				frameSize = (double)sample/samplePerFrame * tmp;
				time++;
				if(fseek(fp,tmp-4,SEEK_CUR) != 0) goto last;
				continue;
			}
			frameSize += tmp;
			sample += samplePerFrame;
			if(fseek(fp,tmp-4,SEEK_CUR) != 0) goto last;
		}
	}
	
  last:
	
	bitrate = round((double)size*8 / (i * (double)samplePerFrame / frequency));
	
	fprintf(fpw1,"set title '%s'\n",argv[1]);
	fprintf(fpw1,"set grid\n");
	if(resolution == frequency) fprintf(fpw1,"set xlabel \"time\"\n");
	else fprintf(fpw1,"set xlabel \"time [sec]\"\n");
	fprintf(fpw1,"set ylabel \"bitrate [bps]\"\n");
	fprintf(fpw1,"average = %d\n",bitrate);
	if(resolution == frequency) {
		fprintf(fpw1,"set xdata time\n");
		fprintf(fpw1,"set timefmt \"%%M:%%S\"\n");
	}
	fprintf(fpw1,"plot '%s' using 1:2 w l title 'bitrate', average\n",filename);
	
  end:
	
	fclose(fpw1);
	fclose(fpw2);
	fclose(fp);
	return 0;
}